package de.popforge.cubicvr.smart3d
{
	import flash.display.BitmapData;
	import flash.geom.Point;
	import flash.geom.Matrix;
	import flash.utils.getTimer;
	
	public class Camera
	{
		static public const LOW: String = 'low';
		static public const MEDIUM: String = 'medium';
		static public const HIGH: String = 'high';
		
		static public const ORIGIN: Point = new Point();
		
		//-- world position and displacement
		public var wx: Number = 0;
		public var wy: Number = 0;
		public var wz: Number = 0;
		public var ax: Number = 0;
		public var ay: Number = 0;
		public var az: Number = 0;
		
		public var eyeAngle: Number;
		public var matrix: Matrix4x3;
		public var inverseMatrix: Matrix4x3;
		
		private var output: BitmapData;
		
		//-- RENDERER
		private var rendererLow: Scanline;
		private var rendererMed: Scanline;
		private var rendererHigh: Scanline;
		private var renderer: Scanline;
		
		public var quality: String;

		private var frustum: Frustum;
		
		private var width: uint;
		private var height: uint;
		private var widthHalf: uint;
		private var heightHalf: uint;
		
		private var zoom: Number;
		private var ratio: Number;
		
		public var useMidMaps: Boolean = false;
				
		public function Camera( width: uint, height: uint, eyeAngle: Number = 128, transparent: Boolean = false )
		{
			this.width = width;
			this.height = height;
			this.eyeAngle = eyeAngle;
			
			widthHalf = width / 2;
			heightHalf = height / 2;
			
			matrix = new Matrix4x3();
			inverseMatrix = new Matrix4x3();
			
			zoom = 1;
			ratio = 1;
			
			frustum = new Frustum( width / zoom, height / zoom, eyeAngle );
			
			rendererLow = new Scanline( width >> 1, height >> 1 );
			rendererMed = new Scanline( width, height );
			rendererHigh = new Scanline( width << 1, height << 1 );
			
			output = new BitmapData( width, height, false, 0 );
			
			setQuality( MEDIUM );
		}
		
		public function setQuality( quality: String ): void
		{
			switch( quality )
			{
				case LOW: renderer = rendererLow; ratio = 2; break;
				case MEDIUM: renderer = rendererMed; ratio = 1; break;
				case HIGH: renderer = rendererHigh; ratio = .5; break;
			}
			
			this.quality = quality;
		}
		
		public function getOutput(): BitmapData
		{
			return output;
		}
		
		public function setZoom( zoom: Number ): void
		{
			if( zoom < 1 ) return;
			
			frustum.setSize( width / zoom, height / zoom );
			
			this.zoom = zoom;
		}
		
		public function getZoom(): Number
		{
			return zoom;
		}
		
		public function setRatio( ratio: Number ): void
		{
			this.ratio = ratio;
		}
		
		public function updateMatrix(): void
		{
			matrix.identity();
			matrix.rotationCAxis( 1, ax );
			matrix.rotationCAxis( 2, ay );
			matrix.rotationCAxis( 3, az );
			matrix.translate( wx, wy, wz );
			
			inverseMatrix = Matrix4x3.inverse( matrix );
		}
		
		public function project( vertex: Vertex ): Boolean
		{
			var tz: Number = vertex.tz;
			
			if( tz <= 0 ) return false;
			
			var pers: Number = eyeAngle / tz;
			vertex.sx = ( vertex.tx * pers * zoom + widthHalf ) / ratio;
			vertex.sy = ( vertex.ty * pers * zoom + heightHalf ) / ratio;
			vertex.sz = 1 / tz;
			
			return true;
		}
				
		public function render( polygons: Array, background: int = 0 ): void
		{
			//-- objects		
			var polygon: Polygon;
			var normal: Vector;
			
			var vertex: Vertex;
			var vertices: Array;
			
			//-- projection
			var pers: Number;
			var tz: Number;
			
			for each( polygon in polygons )
			{
				//-- transform all vertices
				for each( vertex in polygon.pvert )
				{
					inverseMatrix.transformVertex( vertex );
				}
			}
			
			var i: int = polygons.length;
			
			while( --i > -1 )
			{
				polygon = polygons[i];

				//-- fill original vertices to clipped vertices Array (cvert)
				vertices = polygon.cvert = polygon.pvert.slice(0);
	
				//-- Clip Frustrum
				if( vertices.length > 2 ) PolyClipper.clipFrustum( frustum, vertices );
				else
				{
					polygons.splice( i, 1 );
					continue;
				}

				//-- Project to screen if visible
				if( vertices.length > 2 )
				{
					for each( vertex in vertices )
					{
						tz = vertex.tz;
						
						pers = eyeAngle / tz;
						vertex.sx = ( vertex.tx * pers * zoom + widthHalf ) / ratio;
						vertex.sy = ( vertex.ty * pers * zoom + heightHalf ) / ratio;
						vertex.sz = 1 / tz;
					}
				}
				else
				{
					polygons.splice( i, 1 );
					continue;
				}
			}
			
			renderer.clear( background );
			
			for each( polygon in polygons )
			{
				if( polygon.cvert.length > 2 )
				{
					if( useMidMaps )
					{
						polygon.texture = zoom < 2.5 ? polygon.textureBlurred : polygon.textureNormal;
					}
					else
					{
						polygon.texture = polygon.textureNormal;
					}
					
					renderer.drawPolygon( polygon );
				}
			}
			
			switch( quality )
			{
				case LOW:
				case HIGH:
				
					output.draw( renderer.getOutput(), new Matrix( ratio, 0, 0, ratio ), null, null, null, true );
					break;
				
				case MEDIUM:
				
					output.copyPixels( renderer.getOutput(), output.rect, ORIGIN );
					break;
			}
		}
	}
}